from tests.notes_xml import TestNotesXML
from tests.otp_versions_table import TestOTPVersionsTable
from tests.changes import TestChanges

__all__ = ["TestNotesXML", "TestOTPVersionsTable", "TestChanges"]



import argparse
import os.path
import shutil
from six.moves import urllib
import erlang_changes
from .changes import Changes
from .spec import Spec
from .github import get_github_version


ERLANG_SPEC_FILENAME = 'erlang.spec'


try:
	FileNotFoundError
except NameError:
	FileNotFoundError = IOError

class MajorUpdateRequested(Exception):
	def __init__(self, prev_version, version):
		message = "Major update from {} to {} is not supported.".format(prev_version, version)
		super(MajorUpdateRequested, self).__init__(message)

class OTPVersionMismatch(Exception):
	def __init__(self, otp_version, source_otp_version):
		message = "Expected version {}, while {} is found in the source file".format(otp_version, source_otp_version)
		super(OTPVersionMismatch, self).__init__(message)

def get_current_version(filename = ERLANG_SPEC_FILENAME):
	erlang_spec = Spec(filename)
	return erlang_spec.version

def get_spec_sources(filename = ERLANG_SPEC_FILENAME):
	erlang_spec = Spec(filename)
	return erlang_spec.sources

def major_version(version):
	return version.split(".")[0]

parser = argparse.ArgumentParser(prog='otp-service_erlang_changes')
parser.add_argument('--outdir', help='osc service parameter for internal use only', required=True)

def execute_from_commandline(argv=None):
	args = parser.parse_args(argv)

	prev_otp_version = get_current_version()
	prev_otp_major_version = major_version(prev_otp_version)
	otp_version = get_github_version(prev_otp_major_version)

	if prev_otp_version == otp_version:
		return

	if prev_otp_major_version != major_version(otp_version):
		raise MajorUpdateRequested(prev_otp_version, otp_version)

	tmp_spec_filename = os.path.join(args.outdir, ERLANG_SPEC_FILENAME)
	shutil.copyfile(ERLANG_SPEC_FILENAME, tmp_spec_filename)
	Spec.set_version(tmp_spec_filename, otp_version)

	for source in get_spec_sources(tmp_spec_filename):
		url = urllib.parse.urlparse(source)
		if not url.scheme:
			continue
		filename = os.path.basename(urllib.request.url2pathname(url.path))
		if os.path.exists(filename):
			continue
		urllib.request.urlretrieve(source, filename)

	otp_src_filename = "OTP-{version}.tar.gz".format(version = otp_version)
	otp_src = erlang_changes.OTPSrc.from_file(otp_src_filename)

	if otp_version != otp_src.otp_version:
		raise OTPVersionMismatch(otp_version, otp_src.otp_version)

	new_changes = Changes.from_otp_src(otp_src, prev_otp_version)
	for path, changes in Changes.find_changes():
		tmp_changes = os.path.join(args.outdir, changes)
		shutil.copyfile(os.path.join(path, changes), tmp_changes)
		new_changes.write(tmp_changes)

if __name__ == "__main__":
	import sys

	execute_from_commandline(sys.argv[1:])



from collections import OrderedDict
import itertools
import re
from six import viewkeys

class OTPVersionsTable(object):
	_app_re = re.compile(r'(\w+)-((\d+)(\.\d+)*)')

	@staticmethod
	def _proc_otp_version_line(x):
		otp_token, app_token, _ = x.split(":", 2)

		otp_version_match = OTPVersionsTable._app_re.search(otp_token)
		otp_version = otp_version_match.group(2)

		new_apps = itertools.chain(*[x.split() for x in app_token.split("#")])
		apps_version_matches = [OTPVersionsTable._app_re.search(x) for x in new_apps]
		apps = dict([(x.group(1), x.group(2)) for x in apps_version_matches])

		return (otp_version, apps)

	@staticmethod
	def from_file(source):
		if not hasattr(source, "readlines"):
			source = open(source, "r")

		applications = [OTPVersionsTable._proc_otp_version_line(x.decode('ascii')) for x in source.readlines() if x.strip()]

		return OTPVersionsTable(applications)

	def __init__(self, applications):
		versions, self._applications = zip(*applications)
		self._versions = OrderedDict(((ver, idx) for idx, ver in enumerate(versions)))

	@property
	def otp_versions(self):
		return self._versions.keys()

	def __getitem__(self, version):
		return self._applications[self._versions[version]]

	def slice(self, version_begin, version_end):
		idx_begin = self._versions[version_begin] if version_begin else None
		idx_end = self._versions[version_end] if version_end else None

		return itertools.islice(self._versions.keys(), idx_end, idx_begin)

	def diff(self, version):
		idx = self._versions[version]

		apps = self._applications[idx]
		apps_prev = self._applications[idx+1] if (idx+1) < len(self._applications) else dict([])

		keys = viewkeys(apps)
		keys_prev = viewkeys(apps_prev)

		keys_deleted = keys_prev - keys
		keys_added = keys - keys_prev
		keys_common = keys & keys_prev

		changes = ([(k, (apps_prev[k], None)) for k in keys_deleted] + [(k, (None, apps[k])) for k in keys_added] + [(k, (apps_prev[k], apps[k])) for k in keys_common])

		return dict(changes)



import unittest
import io

from erlang_changes import NotesXML

class TestNotesXML(unittest.TestCase):
	def test_parse1(self):
		s = u"""
<chapter>
<section><title>App 1.0.1</title>

    <section><title>Fixed Bugs and Malfunctions</title>
      <list>
        <item>
          <p>Fix <c>1</c></p>
          <p>Skip this</p>
        </item>
        <item>
          <p>Fix <c>2</c></p>
          <p>Another par</p>
          <p>Skip this</p>
        </item>
      </list>
    </section>

    <section><title>Improvements and New Features</title>
      <list>
        <item>
          <p>Feature <c>2</c></p>
          <p>Skip this</p>
        </item>
      </list>
    </section>

</section>

<section><title>App 1.0</title>

    <section><title>Improvements and New Features</title>
      <list>
        <item>
          <p>Feature <c>1</c></p>
          <p>Skip this</p>
        </item>
      </list>
    </section>

</section>

</chapter>
		"""

		slice1_exp = [
			["Fix 1", "Fix 2 Another par", "Feature 2"],
		]
		slice2_exp = [
			["Fix 1", "Fix 2 Another par", "Feature 2"],
			["Feature 1"],
		]

		n = NotesXML.from_xml(io.StringIO(s))

		self.assertListEqual(list(n.app_versions), ["1.0.1", "1.0"])
		self.assertListEqual(n["1.0.1"], ["Fix 1", "Fix 2 Another par", "Feature 2"])
		self.assertListEqual(n["1.0"], ["Feature 1"])
		self.assertListEqual(list(n.slice("1.0", "1.0.1")), slice1_exp)
		self.assertListEqual(list(n.slice(None, "1.0.1")), slice2_exp)
		self.assertListEqual(list(n.slice(None, None)), slice2_exp)
		self.assertListEqual(list(n.slice("1.0", None)), slice1_exp)

if __name__ == '__main__':
	unittest.main()



from erlang_changes.notes_xml import NotesXML
from erlang_changes.otp_src import OTPSrc
from erlang_changes.otp_versions_table import OTPVersionsTable

__all__ = ["NotesXML", "OTPSrc", "OTPVersionsTable"]



from setuptools import setup
from setuptools.command.install import install

OBS_SERVICE_PATH='/usr/lib/obs/service'

class InstallCmd(install, object):
	def finalize_options(self):
		self.install_scripts = OBS_SERVICE_PATH
		super(InstallCmd, self).finalize_options()

setup(name='obs-service-erlang_changes',
	version='0.2.1',
	description='Erlang changelog formatting OBS service',
	url='http://github.com/matwey/obs-service-erlang_changes',
	author='Matwey V. Kornilov',
	author_email='matwey.kornilov@gmail.com',
	license='GPL-2.0',
	packages=['erlang_changes'],
	test_suite='tests',
	entry_points={
		'console_scripts': ['erlang_changes = erlang_changes.cli:execute_from_commandline']
	},
	data_files=[
		(OBS_SERVICE_PATH, ['service/erlang_changes.service']),
	],
	zip_safe=False,
	cmdclass={
		'install': InstallCmd
	}
)



import os.path
import re
import tarfile
import lxml.etree as ET
from erlang_changes.otp_versions_table import OTPVersionsTable
from erlang_changes.notes_xml import NotesXML

class DTDResolver(ET.Resolver):
	dtd_paths = ["lib/erl_docgen/priv/dtd", "lib/erl_docgen/priv/dtd_man_entities"]

	def __init__(self, root_name, otp_src):
		self._root_name = root_name
		self._otp_src = otp_src
		self._cache = dict([])
		super(DTDResolver, self).__init__()

	def _resolve(self, name, no_cache=False):
		entry = self._cache.get(name)

		if not no_cache and entry is not None:
			return entry

		filenames = [os.path.join(self._root_name, p, name) for p in DTDResolver.dtd_paths]
		for x in filenames:
			try:
				entry = self._otp_src.extractfile(x).read()
			except KeyError:
				pass

		if not no_cache and entry is not None:
			self._cache[name] = entry

		return entry

	def resolve(self, url, _id, context):
		name = os.path.basename(url)
		entry = self._resolve(name)

		if entry is not None:
			return self.resolve_string(entry, context)

		return None

class OTPSrc(object):
	_notes_xml_re = re.compile(r'(\w+)/doc/src/notes\.xml$')

	@staticmethod
	def from_file(otp_sources_filename):
		otp_src = tarfile.open(otp_sources_filename)
		filenames = otp_src.getnames()
		root_name = filenames[0]

		otp_version = otp_src.extractfile("{}/OTP_VERSION".format(root_name)).read().decode('ascii').strip()
		otp_versions_table = OTPVersionsTable.from_file(otp_src.extractfile("{}/otp_versions.table".format(root_name)))
		apps = []

		parser = ET.XMLParser(load_dtd=True)
		parser.resolvers.add(DTDResolver(root_name, otp_src))

		for x in filenames:
			app_match = OTPSrc._notes_xml_re.search(x)
			if not app_match:
				continue
			apps.append((app_match.group(1), NotesXML.from_xml(otp_src.extractfile(x), parser=parser)))

		return OTPSrc(otp_version, otp_versions_table, apps)

	def __init__(self, otp_version, otp_versions_table, apps):
		self._otp_version = otp_version
		self._otp_versions_table = otp_versions_table
		self._apps = dict(apps)

	@property
	def otp_version(self):
		return self._otp_version

	@property
	def otp_versions_table(self):
		return self._otp_versions_table

	@property
	def app_notes_xml(self):
		return self._apps



import re
from collections import OrderedDict
import itertools
import lxml.etree as ET

class NotesXML(object):
	_version_re = re.compile(r'((\d+)(\.\d+)*)$')
	_title_list = ['Fixed Bugs and Malfunctions', 'Improvements and New Features', 'Known Bugs and Problems']

	@staticmethod
	def from_xml(source, parser=None):
		if parser is None:
			parser = ET.XMLParser()

		root = ET.parse(source, parser=parser)
		change_list = []

		for x in root.findall('./section[title]'):
			version_match = NotesXML._version_re.search(x.find("title").text)
			if not version_match:
				continue

			version = version_match.group(0)
			entries = []
			for item in itertools.chain.from_iterable([x.findall("./section[title='{}']/list/item[p]".format(title)) for title in NotesXML._title_list]):
				entries.append(" ".join(itertools.chain.from_iterable(["".join(y.itertext()).split() for y in item.xpath("./p[position()<last()]")])))

			change_list.append((version, entries))

		return NotesXML(change_list)

	def __init__(self, changelog):
		versions, self._changes = zip(*changelog)
		self._versions = OrderedDict(((ver, idx) for idx, ver in enumerate(versions)))

	@property
	def app_versions(self):
		return self._versions.keys()

	def __getitem__(self, version):
		return self._changes[self._versions[version]]

	def slice(self, version_begin, version_end):
		idx_begin = self._versions[version_begin] if version_begin else None
		idx_end = self._versions[version_end] if version_end else None

		return itertools.islice(self._changes, idx_end, idx_begin)
